package org.netkernel.client.http.endpoint;

import org.apache.commons.httpclient.HttpState;
import org.apache.commons.httpclient.auth.AuthScope;
import org.netkernel.client.http.representation.HttpStateRepresentation;
import org.netkernel.client.http.representation.IHttpCredentials;
import org.netkernel.client.http.util.GUID;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

/**
 *  An accessor for managing HTTPState objects - stores HTTPConnectionManager, Cookies, credentials etc etc.
 *  Note:  This class is thread-safe but is *not* optimised wrt synchronization around the HashMaps.
 * @author  pjr
 */
public class HttpStateAccessor extends StandardAccessorImpl
{
	private static final HashMap mStateMap=new HashMap();
	private static final HashMap mTimeStamps=new HashMap();
	public static final String ACTIVE_HTTP_STATE="active:httpState";
	public static final long DEFAULT_REAP=600000; //10 minutes
	
	/** Creates a new instance of HTTPStateAccessor */
	public HttpStateAccessor()
	{	this.declareThreadSafe();
		this.declareSupportedVerbs(INKFRequestReadOnly.VERB_SOURCE|INKFRequestReadOnly.VERB_SINK|INKFRequestReadOnly.VERB_DELETE|INKFRequestReadOnly.VERB_EXISTS|INKFRequestReadOnly.VERB_NEW);
	}
	
	@Override
	public void onRequest(INKFRequestContext aContext) throws Exception
	{	if(aContext.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_META)
		{	super.onRequest(aContext);
		}
		else
		{	Object result=null;
			HttpState state=null;
			String stateuri=aContext.getThisRequest().getArgumentValue("id");
			switch(aContext.getThisRequest().getVerb())
			{	case INKFRequestReadOnly.VERB_SOURCE:
					synchronized(mStateMap)
					{	state=(HttpState)mStateMap.get(stateuri);
					}
					if(state==null) throw new NKFException("Error Sourcing HttpState", aContext.getThisRequest().getIdentifier()+" does not exist", null);
					result=new HttpStateRepresentation(state);
					touch(stateuri);
				break;
				case INKFRequestReadOnly.VERB_SINK:
					synchronized(mStateMap)
					{	state=(HttpState)mStateMap.get(stateuri);
						if(state==null) throw new NKFException("Error Sinking HttpState", aContext.getThisRequest().getIdentifier()+" does not exist", null);
						HttpStateRepresentation sa=(HttpStateRepresentation)aContext.sourcePrimary(HttpStateRepresentation.class);
						state=sa.getState();
						mStateMap.put(aContext.getThisRequest().getIdentifier(),state);
					}
					touch(stateuri);
				break;
				case INKFRequestReadOnly.VERB_DELETE:
					synchronized(mStateMap)
					{	state=(HttpState)mStateMap.get(stateuri);
						if(state==null) throw new NKFException("Error Deleting HttpState", aContext.getThisRequest().getIdentifier()+" does not exist", null);
						mStateMap.remove(aContext.getThisRequest().getIdentifier());
					}
				break;
				case INKFRequestReadOnly.VERB_NEW:
					if (stateuri==null)
					{	synchronized(mStateMap)
						{	do
							{	stateuri=GUID.GUID();
							} while(mStateMap.get(stateuri)!=null);
						}
					}
					state=new HttpState();
					//Process Credentials if there are any
					String creduri=aContext.getThisRequest().getArgumentValue("credentials");
					if(creduri!=null)
					{	IHttpCredentials ca=(IHttpCredentials)aContext.source("arg:credentials", IHttpCredentials.class);
						state.setCredentials(new AuthScope(ca.getHost(), ca.getPort(), ca.getRealm()), ca.getCredentials());
					}
					synchronized(mStateMap)
					{	mStateMap.put(stateuri,state);
					}
					StringBuffer sb=new StringBuffer();
					sb.append(ACTIVE_HTTP_STATE);
					sb.append("+id@"+stateuri);
					result=sb.toString();
					touch(stateuri);
				break;
				case INKFRequestReadOnly.VERB_EXISTS:
					synchronized(mStateMap)
					{	state=(HttpState)mStateMap.get(stateuri);
					}
					result=new Boolean(state!=null);
					touch(stateuri);
				break;
				default:
					throw new NKFException("Unknown Request Type");
			}
			//Reap expired state
			reap();
			INKFResponse resp= aContext.createResponseFrom(result);
			/* 
			 Optionally we can set the resource cacheable - but this would double memory for the state!
			if(!(result instanceof VoidAspect))
			{	resp.setCacheable();
			}
			 */
		}
	}
	
	/**
	 * Delete expired state
	 */
	private void reap()
	{	synchronized(mTimeStamps)
		{	synchronized(mStateMap)
			{	Iterator it=mTimeStamps.keySet().iterator();
				ArrayList reapSet=new ArrayList(8);
				while(it.hasNext())
				{	String key=(String)it.next();
					Long timestamp=(Long)mTimeStamps.get(key);
					if(System.currentTimeMillis()>timestamp.longValue())
					{	reapSet.add(key);
					}
				}
				//Clear out the reapable set
				for(int i=0; i<reapSet.size(); i++)
				{	Object key=reapSet.get(i);
					mStateMap.remove(key);
					mTimeStamps.remove(key);
				}
			}
		}
	}
	
	/**
	 * Touch state which has been interacted with so that it has a new expiry lease.
	 */
	private void touch(String stateuri)
	{	synchronized(mTimeStamps)
		{	mTimeStamps.put(stateuri,new Long(System.currentTimeMillis()+DEFAULT_REAP));
		}
	}
}